var MetaDataFactory = function(localStorageEngine, requestDataSourceEngine, requestTotalizerDataSourceEngine, SQLStorageStrategy, restEngine, eventEngine,
	ScreenService, aclEngine, templateManager, ZhUtil, ZHPromise, dataSourceHolder, ZhLogger, DecodeValueFieldService, ZhFeatures, memoryStorageEngine,
	validateEngine, MessageService) {

	var self = this;
	var _extends = Util.extend;

	var _sqlDataSourceFactory = function(obj) {
		var fileName = obj.name.replace("/", "") + ".json";
		restEngine.requestMetaData(Util.getRightPath(fileName, "json/datasources/"), {}, function(data) {
			SQLStorageStrategy.registerDataSource(data);
		});
		return _defaultDataSourceFactory(obj, SQLStorageStrategy);
	};

	var _restDataSourceFactory = function(obj) {
		var dataSource = _defaultDataSourceFactory(obj, requestDataSourceEngine);
		dataSource.name = dataSource.serviceName || dataSource.name;
		return dataSource;
	};

	var _totalizerDataSourceFactory = function(obj) {
		var dataSource = _defaultDataSourceFactory(obj, requestTotalizerDataSourceEngine);
		dataSource.name = dataSource.serviceName || dataSource.name;
		return dataSource;
	};

	var _localStorageDataSourceFactory = function(obj) {
		return _defaultDataSourceFactory(obj, localStorageEngine);
	};

	var _memoryStorageDataSourceFactory = function(obj) {
		return _defaultDataSourceFactory(obj, memoryStorageEngine);
	};

	var _defaultDataSourceFactory = function(obj, dataSourceEngine) {
		return _extends(new DataSource(dataSourceEngine, templateManager, eventEngine, dataSourceHolder, ZHPromise, ScreenService), obj);
	};

	var _loadDataSourceData = function(dataSource, owner, filter) {
		if ((!dataSource.data || !dataSource.data.length) && !dataSource.lazyLoad) {
			dataSource.load(filter, 1, owner.itemsPerPage, undefined, undefined, undefined, undefined, undefined, owner).then(function() {
				if (dataSource.onInit) {
					dataSource.onInit();
				}
			});
		} else if (dataSource.onInit) {
			dataSource.onInit();
		}
	};

	var _dataSourceFactory = function(obj, filter, owner) {
		var dataSource = obj;
		if (!(obj instanceof DataSource)) {
			filter = filter || [];
			owner = owner || {};
			var hasStorageStrategy = obj.localStorage || obj.rest || obj.sql || obj.memoryStorage || obj.totalizer;
			if (hasStorageStrategy) {
				if (obj.sql) {
					dataSource = _sqlDataSourceFactory(obj);
				} else if (obj.rest) {
					dataSource = _restDataSourceFactory(obj);
				} else if(obj.memoryStorage) {
					dataSource = _memoryStorageDataSourceFactory(obj);
				} else if(obj.totalizer) {
					dataSource = _totalizerDataSourceFactory(obj);
				} else {
					dataSource = _localStorageDataSourceFactory(obj);
				}
				_loadDataSourceData(dataSource, owner, filter );
			} else {
				dataSource = _defaultDataSourceFactory(obj);
			}
			dataSource.owner = owner;
		}
		return dataSource;
	};

	this.dataSourceFactory = _dataSourceFactory;
	this.factoryDataSource = _dataSourceFactory;

	this.fieldFactory = function(obj, widget, appendToWidget) {
		var field = _extends(new Field(ZhUtil, templateManager, DecodeValueFieldService, validateEngine), obj);
		registerFieldFunctionsUsage(obj, field);
		_extendMobileProperties(field);
		if (field.dataSource) {
			field.dataSource = _dataSourceFactory(field.dataSource, field.dataSourceFilter, field);
		}
		if (widget) {
			field.widget = widget;
			field.setSequence(widget.getNextFieldSequence());
			if (appendToWidget) {
				widget.fields.push(field);
			}
		}
		return field;
	};
	/**
	 * Feature tracking. Muste be removed after we find out if these features are used or not
	 */
	function registerFieldFunctionsUsage(baseObject, field) {
		// Repeats everything instead of making it modular to maintain the calee name
		if (!baseObject.getDataSource || isNullFunction(baseObject.getDataSource.toString())) {
			field.getDataSource = function getDataSource() {
				ZhFeatures.registerFeatureUsage({fieldName: field.name});
			};
		}

		if (!baseObject.show || isNullFunction(baseObject.show.toString())) {
			field.show = function show() {
				ZhFeatures.registerFeatureUsage({fieldName: field.name});
			};
		}

		if (!baseObject.hide || isNullFunction(baseObject.hide.toString())) {
			field.hide = function hide() {
				ZhFeatures.registerFeatureUsage({fieldName: field.name});
			};
		}
	}

	function isNullFunction(functionCode) {
		return functionCode.replace(/ /g, '') === "function(){}";
	}

	this.factoryField = this.fieldFactory;

	var _prepareFields = function(widget) {
		if (!Array.isArray(widget.fields)) {
			throw 'ZEEDHI ERROR:\nThe property "fields" of a widget is required to be an Array:\nAt Widget "' +
			widget.name + '"';
		}
		widget.fields = widget.fields.map(function(field) {
			return self.fieldFactory(field, widget);
		});
	};

	function _prepareFieldGroups(widget) {
		if (widget.fieldGroups) {
			widget.fieldGroups = widget.fieldGroups.map(function(item) {
				item.widget = widget;
				return Util.extend(new FieldGroup(), item);
			});
		}
	}

	this.actionFactory = function(obj, widget, appendToWidget) {
		var action = _extends(new Action(), obj);
		_extendMobileProperties(action);
		if (widget) {
			action.widget = widget;

			if (appendToWidget) {
				widget.actions.push(action);
			}

			if (widget.container) {
				widget.container.actions = widget.container.actions || [];
				widget.container.actions.push(action);
			}
		}
		return action;
	};

	function _prepareActions(widget) {
		if (widget.actions) {
			widget.actions = widget.actions.map(function(item) {
				item.widget = widget;
				var action = Util.extend(new Action(), item);
				_extendMobileProperties(action);
				return action;
			});
		}
	}

	function _prepareDataSource(widget, initialFilter) {
		if (widget.dataSource) {
			var filters = Array.isArray(initialFilter) ? widget.getDataSourceFilter().concat(initialFilter) : widget.getDataSourceFilter();
			widget.dataSource = _dataSourceFactory(widget.dataSource, filters, widget);
			var dataSource = widget.dataSource;
			dataSource.onInit = function() {
				if (widget.initDataSource) {
					widget.initDataSource();
				}
			};
			if (widget.afterLoadDataSource) {
				dataSource.afterLoadDataSource(widget.afterLoadDataSource);
			}
			if (widget.afterLoadDataSourceError) {
				dataSource.afterLoadDataSourceError(widget.afterLoadDataSourceError);
			}
		}
	}

	function _prepareWidgetEvents(widget) {
		var events = widget.events;
		if (events) {
			eventEngine.bindEvents(widget, events, function() {
				return widget.dataSource ? widget.dataSource.data : [];
			}, function() {
				return widget.currentRow ? widget.currentRow : {};
			}, function() {
				return widget.selectedRow ? widget.selectedRow : {};
			}, function() {
				return widget.rowGroup ? widget.rowGroup : {};
			});
		}
	}

	function _prepareEvents(owner) {
		var events = owner.events;
		if (events) {
			eventEngine.bindEvents(owner, events);
		}
	}

	function _makeFilter(name, widget, refName) {
		return {
			"name": name,
			"value": function() {
				return widget.currentRow ? widget.currentRow[refName] : undefined;
			},
			"operator": "="
		};
	}

	function _prepareDefaultValue(widget, innerWidget) {
		var parentFilter = innerWidget.parentFilter;
		if (parentFilter && parentFilter.refkeys && parentFilter.refkeys.length && parentFilter.parentkeys &&
			parentFilter.parentkeys.length) {
			parentFilter.refkeys.forEach(function(name, i) {
				var refName = parentFilter.parentkeys[i];
				var field = innerWidget.getField(name);
				field.defaultValue = function() {
					return widget.currentRow[refName];
				};
			});
		}
	}

	function _prepareParentFilter(widget, innerWidget) {
		var parentFilter = innerWidget.parentFilter;
		if (parentFilter && parentFilter.refkeys && parentFilter.refkeys.length && parentFilter.parentkeys &&
			parentFilter.parentkeys.length) {
			parentFilter.refkeys.forEach(function(name, i) {
				var refName = parentFilter.parentkeys[i];
				innerWidget.dataSourceFilter.push(_makeFilter(name, widget, refName));
			});
		}
	}

	function _prepareChildren(widget, shouldAddToFactoredWidgets) {
		if (widget.widgets) {
			widget.widgets = widget.widgets.map(function(innerWidget) {
				var preparedWidget = _factoryWidget(innerWidget, undefined, shouldAddToFactoredWidgets);
				preparedWidget.parent = widget;
				_prepareParentFilter(widget, preparedWidget);
				_prepareDefaultValue(widget, preparedWidget);
				_prepareEventUpdateDataSource(preparedWidget);
				return preparedWidget;
			});
		}
	}

	function _prepareEventUpdateDataSource(preparedWidget) {
		if (!preparedWidget.onEnterTab) {
			eventEngine.createEvent(preparedWidget, "onEnterTab", true);
		}
		preparedWidget.onEnterTab(function(event) {
			if (!event.isPropagationStopped() && widgetIsApplicableToReloadOnEnterTab(preparedWidget)) {
				preparedWidget.reload();
			}
		});
	}

	function widgetIsApplicableToReloadOnEnterTab(widget) {
		var isMasterDetail = isAnyMasterDetailTemplate(widget);
		var isMasterDetailOnList = isMasterDetail && widget.mode === ViewMode.LIST;
		var widgetIsPreventingReload = widget.isLazyLoad() || !widget.reloadOnEnterTab;
		return (!isMasterDetail || isMasterDetailOnList) && !widgetIsPreventingReload;
	}

	function isAnyMasterDetailTemplate(widget) {
		var isMasterDetailGrid = widget.template && ~widget.template.indexOf("master_detail/grid.html");
		var isMasterDetailList = widget.template && ~widget.template.indexOf("master_detail/list.html");
		var isMasterDetailTreeGrid = widget.template && ~widget.template.indexOf("master_detail/tree/grid.html");
		return isMasterDetailGrid || isMasterDetailList || isMasterDetailTreeGrid;
	}

	var _validateWidget = function(obj) {
		if (!obj.container) {
			throw new Error("Widget should have a container!");
		}
	};

	/**
	 * Extends or replaces each property of the object with the ones specified in obj.mobile
	 *
	 * @param {object} obj the object to search for mobile properties
	 */
	function _extendMobileProperties (obj) {
		if (Util.isMobile() && obj.mobile) {
			var mobile = obj.mobile;
			Object.keys(mobile).forEach(function extendMobileProperty (propertyToExtend) {
				if (obj.replaceByMobile || !obj[propertyToExtend] || typeof obj[propertyToExtend] !== 'object') {
					obj[propertyToExtend] = mobile[propertyToExtend];
				} else {
					obj[propertyToExtend] = Util.extend(obj[propertyToExtend], mobile[propertyToExtend]);
				}
			});
		}
	}

	function _addToFactoredWidgets (widget, shouldAddToFactoredWidgets) {
		if (shouldAddToFactoredWidgets) {
			widget.container.factoredWidgets = widget.container.factoredWidgets || [];
			widget.container.factoredWidgets.push(widget);
		}
	}

	var _factoryWidget = function(obj, initialFilter, shouldAddToFactoredWidgets) {
		_validateWidget(obj);
		var widget = _extends(new Widget(ScreenService, ZhUtil, eventEngine, ZHPromise, templateManager, ZhLogger, ZhFeatures, MessageService), obj);
		_addToFactoredWidgets(widget, shouldAddToFactoredWidgets);
		_extendMobileProperties(widget);
		_prepareWidgetEvents(widget);
		_prepareFieldGroups(widget);
		_prepareFields(widget);
		_prepareActions(widget);
		_prepareDataSource(widget, initialFilter);
		_prepareChildren(widget, shouldAddToFactoredWidgets);
		return widget;
	};

	var _setContainer = function(container, item) {
		item.container = container;
		if (item.widgets) {
			item.widgets.forEach(function(item) {
				_setContainer(container, item);
			});
		}
	};
	var _factoryWidgets = function(container, maintainWindowFilter, shouldAddToFactoredWidgets) {
		container.widgets.forEach(function(widget) {
			_setContainer(container, widget);
		});
		container.widgets = container.widgets.map(function(widget) {
			return _factoryWidget(widget, Util.isUndefinedOrNull(maintainWindowFilter) ? null : maintainWindowFilter[widget.name], shouldAddToFactoredWidgets);
		});
	};

	var _factoryNotifications = function _factoryNotifications(project) {
		if (project && project.notifications) {
			project.notifications = project.notifications.map(function (notification) {
				notification = _extends(new _Notification(), notification);
				_extendMobileProperties(notification);
				return notification;
			});
		}
	};

	var _activeFirstWidgets = function(container) {
		container.widgets[0].activate();
	};

	this.processContainerAcl = function(container) {
		if (templateManager.project && templateManager.project.aclEnable && container.aclEnable !==
			false) {
			aclEngine.processAcl(container, templateManager.aclRules, templateManager.userId, templateManager.project.aclDefaultRule);
			container.aclLoaded = true;
		}
	};

	var _processProjectAcl = function _processProjectAcl(project) {
		if (project && project.aclEnable) {
			aclEngine.processAcl(project, templateManager.aclRules, templateManager.userId, project.aclDefaultRule);
		}
	};

	this.containerFactory = function(obj, maintainWindowFilter, shouldAddToFactoredWidgets) {
		var container;
		if (!obj.$initialized) {
			container = _extends(new Container(ScreenService, MessageService), obj);
			_extendMobileProperties(container);
			container.maintainWindowFilter = maintainWindowFilter;
			_factoryWidgets(container, maintainWindowFilter, shouldAddToFactoredWidgets);
			container.$initialized = true;
		} else {
			container = obj;
			container.reloadLazyDataSources(maintainWindowFilter);
		}
		_activeFirstWidgets(container);
		this.processContainerAcl(container);
		_prepareEvents(container);
		return container;
	};

	this.wizardFactory = function(obj) {
		var wizard = _extends(new Wizard(), obj);
		wizard.init();
		return wizard;
	};

	this.widgetFactory = function(widget, container) {
		if (container instanceof Container) {
			_setContainer(container, widget);
		}
		return _factoryWidget(widget, undefined, true);
	};

	this.menuFactory = function(menu, parent) {
		menu = _extends(new Menu(), menu);
		menu.parent = parent;
		_prepareEvents(menu);
		if (menu.menus && menu.menus.length > 0) {
			menu.menus = self.groupMenuFactory(menu.menus, menu);
		}
		return menu;
	};

	this.groupMenuFactory = function(groupMenu, parent) {
		if (groupMenu && groupMenu.length > 0) {
			groupMenu = groupMenu.map(function(menu) {
				return self.menuFactory(menu, parent);
			});
		}
		return groupMenu;
	};

	this.projectFactory = function(obj) {
		var project;
		if (!obj.$initialized) {
			project = _extends(new Project(this, templateManager), obj);
			_extendMobileProperties(project);
			_factoryNotifications(project);
			project.groupMenu = this.groupMenuFactory(project.groupMenu);
			_processProjectAcl(project);
			project.$initialized = true;
			project.zhFeatureUsageEnabled = getZhFeatureUsageEnabledAccordingToProjectDefinition(obj, project);

		} else {
			project = obj;
		}
		return project;
	};

	function getZhFeatureUsageEnabledAccordingToProjectDefinition(project, factoredProject) {
		return Util.isUndefined(project.zhFeatureUsageEnabled)? factoredProject.inDeveloperMode : factoredProject.zhFeatureUsageEnabled;
	}
};